from __future__ import annotations

import random
from pathlib import Path
from typing import Optional

from rdkit import Chem

# Basic SMILES that contain different functional groups
BASIC_SMILES = [
    "CC=C",  # alkene,
    "CC#C",  # alkyne,
    "CC[OH]",  # alcohol
    "COC",  # ether
    "CC(=O)[OH]",  # carboxylic acid
    "CC=O",  # aldehyde
    "CC(=O)C",  # ketone
    "CC(=O)N",  # amide
    "CCN",  # amine
    "CC#N",  # nitrile
]


def random_zinc(size: int, rng: Optional[random.Random] = None) -> list[str]:
    """Return random SMILES from the ZINC250k dataset."""
    zinc_path = Path(__file__).parent / "data" / "zinc250k.smiles"
    assert zinc_path.exists() and zinc_path.is_file()
    with open(zinc_path, "r") as f:
        all_zinc_smiles = f.readlines()

    # Remove whitespace and remove empty strings
    all_zinc_smiles = [s.strip() for s in all_zinc_smiles]
    all_zinc_smiles = [s for s in all_zinc_smiles if s]

    # Sample random SMILES
    rng = rng or random.Random()
    chosen_smiles = rng.choices(population=all_zinc_smiles, k=size)

    # Last check: ensure that all SMILES are canonical
    canon_smiles = [Chem.CanonSmiles(s) for s in chosen_smiles]
    return [s for s in canon_smiles if s]  # exclude None



from __future__ import annotations

import math

import joblib
import pytest
from rdkit import Chem
from rdkit.Chem import QED, Crippen

from mol_ga import default_ga
from mol_ga.mol_libraries import random_zinc


def qed_value(smiles: str) -> float:
    """Calculate QED value of a SMILES string."""
    mol = Chem.MolFromSmiles(smiles)
    if mol is None:
        return -1.0
    return QED.qed(mol)


def batch_qed(smiles: list[str]) -> list[float]:
    """Calculate QED values of a list of SMILES strings."""
    return [qed_value(s) for s in smiles]


def qed_logp_gmean(smiles: str) -> float:
    """Geometric mean of qed and logp, negative if logp is negative."""
    mol = Chem.MolFromSmiles(smiles)
    if mol is None:
        return -1e9
    logp_val = Crippen.MolLogP(mol)
    qed_val = QED.qed(mol)
    gmean = math.sqrt(abs(logp_val) * qed_val)
    if logp_val < 0:
        return -gmean
    else:
        return gmean


def batch_qed_logp_gmean(smiles: list[str]) -> list[float]:
    return [qed_logp_gmean(s) for s in smiles]


@pytest.mark.parametrize("parallel", [False, True])
def test_smoke(parallel: bool):
    """Test that the default GA runs with and without parallel."""
    kwargs = dict(
        starting_population_smiles=random_zinc(1000),
        scoring_function=batch_qed,
        max_generations=10,
        offspring_size=100,
    )

    if parallel:
        with joblib.Parallel(n_jobs=-1) as parallel:
            output = default_ga(**kwargs, parallel=parallel)
    else:
        output = default_ga(**kwargs, parallel=None)

    assert max(output.population)[0] > 0.935


def test_improvement():
    """
    Test that the GA sufficiently improves an objective which is easy to
    optimize (combination of QED and logP)

    It should be slightly harder to optimize this than to purely optimize logP,
    but still easy enough to see quick improvement.
    """
    start_population = random_zinc(1000)
    output = default_ga(
        starting_population_smiles=list(start_population),
        scoring_function=batch_qed_logp_gmean,
        max_generations=50,
        offspring_size=25,
    )

    # Test is improvement over starting population.
    # Improvement should be at least over a small threshold
    best_value = max(output.population)[0]
    best_starting_value = max(output.scoring_func_evals[s] for s in start_population)
    assert best_value - best_starting_value >= 0.05  # small but non-trivial improvement


def test_empty_population():
    """
    Test that passing an empty population raises an error.
    """
    with pytest.raises(ValueError):
        default_ga(
            starting_population_smiles=[],
            scoring_function=batch_qed,
            max_generations=10,
            offspring_size=100,
        )



from __future__ import annotations

from random import Random
from typing import Optional

import joblib
from rdkit import Chem, RDLogger

from . import crossover as co
from . import mutate as mu

rd_logger = RDLogger.logger()


def sanitize_mol_to_smiles(mol: Chem.Mol) -> Optional[str]:
    try:
        Chem.SanitizeMol(mol)
        return Chem.MolToSmiles(mol, canonical=True, isomericSmiles=True)
    except Exception:  # unsure what exceptions are thrown
        return None


def reproduce(
    smiles1: str, smiles2: str, mutation_rate: float, rng: Random, crossover_kwargs: Optional[dict] = None
) -> Optional[str]:
    """

    Args:
        mating_pool: list of RDKit Mol
        mutation_rate: rate of mutation

    Returns:

    """

    # turn off rdkit logging
    rd_logger.setLevel(RDLogger.CRITICAL)

    parent_a = Chem.MolFromSmiles(smiles1)
    parent_b = Chem.MolFromSmiles(smiles2)
    crossover_kwargs = crossover_kwargs or dict()
    try:
        new_child = co.crossover(parent_a, parent_b, rng, **crossover_kwargs)
    except RuntimeError:
        return None  # occasionally this happens due to internal rdkit errors
    if new_child is not None and rng.random() < mutation_rate:
        new_child = mu.mutate(new_child, rng)
    if new_child is None:
        return None
    else:
        return sanitize_mol_to_smiles(new_child)


def mutate(smiles: str, rng: Random) -> Optional[str]:
    """Performs Graph GA mutations on a SMILES string"""
    mol = Chem.MolFromSmiles(smiles)
    new_mol = mu.mutate(mol, rng)
    if new_mol is None:
        return None
    else:
        return sanitize_mol_to_smiles(new_mol)


def graph_ga_blended_generation(
    samples: list[str],
    n_candidates: int,
    rng: Random,
    parallel: Optional[joblib.Parallel] = None,
    frac_graph_ga_mutate: float = 0.10,
) -> set[str]:
    """
    Generate candidates with a blend between Graph GA crossover (with some mutation)
    and Graph GA mutate only.
    """

    # Turn off logging
    rd_logger = RDLogger.logger()
    rd_logger.setLevel(RDLogger.CRITICAL)

    # Step 1: divide samples into "mutate" and "reproduce" sets
    samples_mutate: list[str] = []
    samples_crossover: list[str] = []
    for s in samples:
        if rng.random() < frac_graph_ga_mutate:
            samples_mutate.append(s)
        else:
            samples_crossover.append(s)
    # Ensure there are not too many samples in the mutate set
    samples_mutate = samples_mutate[: int(n_candidates * frac_graph_ga_mutate + 1)]  # add one to avoid rounding errors

    # Step 2: run mutations
    if parallel:
        offspring = parallel(joblib.delayed(mutate)(s, rng) for s in samples_mutate)
    else:
        offspring = [mutate(s, rng) for s in samples_mutate]

    # Step 3: run crossover betweeen the crossover samples and a shuffled version of itself
    n_crossover = n_candidates - len(offspring)
    crossover_pairs = list(samples_crossover)
    rng.shuffle(crossover_pairs)
    crossover_mut_rate = 1e-2
    if parallel:
        offspring += parallel(
            joblib.delayed(reproduce)(s1, s2, crossover_mut_rate, rng)
            for s1, s2 in zip(samples_crossover[:n_crossover], crossover_pairs)
        )
    else:
        offspring += [
            reproduce(s1, s2, crossover_mut_rate, rng)
            for s1, s2 in zip(samples_crossover[:n_crossover], crossover_pairs)
        ]

    # Step 4: post-process and return offspring
    offspring = set(offspring)
    offspring.discard(None)  # this sometimes is returned
    return offspring
